﻿namespace NHibernateTest
{
    using NHibernate.Engine;
    using NHibernate.Proxy;
    using NHibernate.Type;

    /// <summary>
    /// Ez az osztály arra a NHibernate hibára nyújt javítást, hogy olyan entitás törlésekor, ami
    /// tartalmaz nem inicialitált lazy komponenst, ClassCastException dobódik, mivel mindenáron
    /// ellenőrizni akarja a törlés során a nem inicializált (és amúgy nem szükséges) mezőket.
    /// Fontos: az osztály a Hibernate 3.3.3.4000-es verziójából készült, így a bináris assembly
    /// frissítésekor ezt az osztályt is frissíteni kell, amennyiben változott az eredeti.
    /// 
    /// Lásd még: <see cref="ComponentSafeDeleteEventListener"/>
    /// </summary>
    public class ComponentSafeNullifier
    {
        private readonly bool isDelete;
        private readonly bool isEarlyInsert;
        private readonly ISessionImplementor session;
        private readonly object self;

        public ComponentSafeNullifier(object self, bool isDelete, bool isEarlyInsert, ISessionImplementor session)
        {
            this.isDelete = isDelete;
            this.isEarlyInsert = isEarlyInsert;
            this.session = session;
            this.self = self;
        }

        /// <summary> 
        /// Nullify all references to entities that have not yet 
        /// been inserted in the database, where the foreign key
        /// points toward that entity
        /// </summary>
        public void NullifyTransientReferences(object[] values, IType[] types)
        {
            for (int i = 0; i < types.Length; i++)
            {
                values[i] = NullifyTransientReferences(values[i], types[i]);
            }
        }

        /// <summary> 
        /// Return null if the argument is an "unsaved" entity (ie. 
        /// one with no existing database row), or the input argument 
        /// otherwise. This is how Hibernate avoids foreign key constraint
        /// violations.
        /// </summary>
        private object NullifyTransientReferences(object value, IType type)
        {
            if (value == null)
            {
                return null;
            }
            else if (type.IsEntityType)
            {
                EntityType entityType = (EntityType)type;
                if (entityType.IsOneToOne)
                {
                    return value;
                }
                else
                {
                    string entityName = entityType.GetAssociatedEntityName();
                    return IsNullifiable(entityName, value) ? null : value;
                }
            }
            else if (type.IsAnyType)
            {
                return IsNullifiable(null, value) ? null : value;
            }
            else if (type.IsComponentType)
            {
                IAbstractComponentType actype = (IAbstractComponentType)type;

                // in case of lazy component, before we procees, let's check
                // if there are any subtypes that might be interested in nullifying
                if (value is NHibernate.Intercept.UnfetchedLazyProperty)
                {
                    // TODO: check if component contains properties which might be nullified regardless of lazy or not
                    return value;
                }

                object[] subvalues = actype.GetPropertyValues(value, session);
                IType[] subtypes = actype.Subtypes;
                bool substitute = false;
                for (int i = 0; i < subvalues.Length; i++)
                {
                    object replacement = NullifyTransientReferences(subvalues[i], subtypes[i]);
                    if (replacement != subvalues[i])
                    {
                        substitute = true;
                        subvalues[i] = replacement;
                    }
                }
                if (substitute)
                    actype.SetPropertyValues(value, subvalues, session.EntityMode);
                return value;
            }
            else
            {
                return value;
            }
        }

        /// <summary> 
        /// Determine if the object already exists in the database, using a "best guess"
        /// </summary>
        private bool IsNullifiable(string entityName, object obj)
        {

            //if (obj == org.hibernate.intercept.LazyPropertyInitializer_Fields.UNFETCHED_PROPERTY)
            //  return false; //this is kinda the best we can do...

            if (obj.IsProxy())
            {
                INHibernateProxy proxy = obj as INHibernateProxy;

                // if its an uninitialized proxy it can't be transient
                ILazyInitializer li = proxy.HibernateLazyInitializer;
                if (li.GetImplementation(session) == null)
                {
                    return false;
                    // ie. we never have to null out a reference to
                    // an uninitialized proxy
                }
                else
                {
                    //unwrap it
                    obj = li.GetImplementation();
                }
            }

            // if it was a reference to self, don't need to nullify
            // unless we are using native id generation, in which
            // case we definitely need to nullify
            if (obj == self)
            {
                // TODO H3.2: Different behaviour
                //return isEarlyInsert || (isDelete && session.Factory.Dialect.HasSelfReferentialForeignKeyBug);
                return isEarlyInsert || isDelete;
            }

            // See if the entity is already bound to this session, if not look at the
            // entity identifier and assume that the entity is persistent if the
            // id is not "unsaved" (that is, we rely on foreign keys to keep
            // database integrity)
            EntityEntry entityEntry = session.PersistenceContext.GetEntry(obj);
            if (entityEntry == null)
            {
                return ForeignKeys.IsTransient(entityName, obj, null, session);
            }
            else
            {
                return entityEntry.IsNullifiable(isEarlyInsert, session);
            }
        }
    }
}
